@@ -7,11 +7,15 @@ Created on 27.09.2017
 
 import asyncio
 import enum
+from html import unescape
+import json
 import logging
-import re
+from typing import Any, Dict
 
 import aiohttp
 
+from .const import INVERTER_DEVICE_TYPE
+
 _LOGGER = logging.getLogger(__name__)
 DEGREE_CELSIUS = "°C"
 WATT = "W"
@@ -36,7 +40,7 @@ class API_VERSION(enum.Enum):
 
 API_BASEPATHS = {
     API_VERSION.V0: "/solar_api/",
-    API_VERSION.V1: "/solar_api/v1",
+    API_VERSION.V1: "/solar_api/v1/",
 }
 
 URL_API_VERSION = "solar_api/GetAPIVersion.cgi"
@@ -47,6 +51,9 @@ URL_SYSTEM_INVERTER = {
     API_VERSION.V1: "GetInverterRealtimeData.cgi?Scope=System",
 }
 URL_SYSTEM_LED = {API_VERSION.V1: "GetLoggerLEDInfo.cgi"}
+URL_SYSTEM_STORAGE = {
+    API_VERSION.V1: "GetStorageRealtimeData.cgi?Scope=System"
+}
 URL_DEVICE_METER = {API_VERSION.V1: "GetMeterRealtimeData.cgi?Scope=Device&DeviceId={}"}
 URL_DEVICE_STORAGE = {
     API_VERSION.V1: "GetStorageRealtimeData.cgi?Scope=Device&DeviceId={}"
@@ -78,18 +85,76 @@ URL_DEVICE_INVERTER_COMMON = {
 URL_ACTIVE_DEVICE_INFO_SYSTEM = {
     API_VERSION.V1: "GetActiveDeviceInfo.cgi?DeviceClass=System"
 }
+URL_INVERTER_INFO = {
+    API_VERSION.V0: "GetInverterInfo.cgi",
+    API_VERSION.V1: "GetInverterInfo.cgi",
+}
 URL_LOGGER_INFO = {
     API_VERSION.V0: "GetLoggerInfo.cgi",
     API_VERSION.V1: "GetLoggerInfo.cgi",
 }
 
+HEADER_STATUS_CODES = {
+    0: "OKAY",
+    1: "NotImplemented",
+    2: "Uninitialized",
+    3: "Initialized",
+    4: "Running",
+    5: "Timeout",
+    6: "Argument Error",
+    7: "LNRequestError",
+    8: "LNRequestTimeout",
+    9: "LNParseError",
+    10: "ConfigIOError",
+    11: "NotSupported",
+    12: "DeviceNotAvailable",
+    255: "UnknownError",
+}
 
-class NotSupportedError(ValueError):
+
+class FroniusError(Exception):
     """
-    An error to be raised if a specific feature is not supported by the specified device
+    A superclass that covers all errors occuring during the
+    connection to a Fronius device
     """
 
-    pass
+
+class NotSupportedError(ValueError, FroniusError):
+    """
+    An error to be raised if a specific feature
+    is not supported by the specified device
+    """
+
+
+class ConnectionError(ConnectionError, FroniusError):
+    """
+    An error to be raised if the connection to the fronius device failed
+    """
+
+
+class InvalidAnswerError(ValueError, FroniusError):
+    """
+    An error to be raised if the host Fronius device could not answer a request
+    """
+
+
+class BadStatusError(FroniusError):
+    """A bad status code was returned."""
+    def __init__(
+            self,
+            endpoint: str,
+            code: int,
+            reason: str = None,
+            response: Dict[str, Any] = {},
+            ) -> None:
+        """Instantiate exception."""
+        self.response = response
+        message = (
+            f"BadStatusError at {endpoint}. "
+            f"Code: {code} - {HEADER_STATUS_CODES.get(code, 'unknown status code')}. "
+            f"Reason: {reason or 'unknown'}."
+        )
+        super().__init__(message)
 
 
 class Fronius:
@@ -110,12 +175,14 @@ class Fronius:
         Constructor
         """
         self._aio_session = session
-        self.url = re.sub(r"/+$", "", url)
+        while url[-1] == "/":
+            url = url[:-1]
+        self.url = url
         # prepend http:// if missing, by fronius API this is the only supported protocol
         if not self.url.startswith("http"):
             self.url = "http://{}".format(self.url)
         self.api_version = api_version
-        self.base_url = API_BASEPATHS.get(API_VERSION)
+        self.base_url = API_BASEPATHS.get(api_version)
 
     async def _fetch_json(self, url):
         """
@@ -124,7 +191,7 @@ class Fronius:
         try:
             async with self._aio_session.get(url) as res:
                 result = await res.json(content_type=None)
-        except (aiohttp.ServerTimeoutError, asyncio.TimeoutError):
+        except asyncio.TimeoutError:
             raise ConnectionError(
                 "Connection to Fronius device timed out at {}.".format(url)
             )
@@ -132,8 +199,10 @@ class Fronius:
             raise ConnectionError(
                 "Connection to Fronius device failed at {}.".format(url)
             )
-        except aiohttp.ContentTypeError:
-            raise ValueError("Host returned a non-JSON reply at {}.".format(url))
+        except (aiohttp.ContentTypeError, json.decoder.JSONDecodeError):
+            raise InvalidAnswerError(
+                "Host returned a non-JSON reply at {}.".format(url)
+            )
         return result
 
     async def fetch_api_version(self):
@@ -144,7 +213,7 @@ class Fronius:
         try:
             res = await self._fetch_json("{}/{}".format(self.url, URL_API_VERSION))
             api_version, base_url = API_VERSION(res["APIVersion"]), res["BaseURL"]
-        except ValueError:
+        except InvalidAnswerError:
             # Host returns 404 response if API version is 0
             api_version, base_url = API_VERSION.V0, API_BASEPATHS[API_VERSION.V0]
 
@@ -176,12 +245,11 @@ class Fronius:
                 )
         spec_url = spec.get(self.api_version)
         if spec_url is None:
-            _LOGGER.warning(
+            raise NotSupportedError(
                 "API version {} does not support request of {} data".format(
                     self.api_version, spec_name
                 )
             )
-            return None
         if spec_formattings:
             spec_url = spec_url.format(*spec_formattings)
 
@@ -192,19 +260,23 @@ class Fronius:
     async def fetch(
         self,
         active_device_info=True,
+        inverter_info=True,
         logger_info=True,
         power_flow=True,
         system_meter=True,
         system_inverter=True,
-        device_meter=frozenset([0]),
+        system_storage=True,
+        device_meter=frozenset(["0"]),
         # storage is not necessarily supported by every fronius device
-        device_storage=frozenset([0]),
-        device_inverter=frozenset([1]),
+        device_storage=frozenset(["0"]),
+        device_inverter=frozenset(["1"]),
         loop=None,
     ):
         requests = []
         if active_device_info:
             requests.append(self.current_active_device_info())
+        if inverter_info:
+            requests.append(self.inverter_info())
         if logger_info:
             requests.append(self.current_logger_info())
         if power_flow:
@@ -213,6 +285,8 @@ class Fronius:
             requests.append(self.current_system_meter_data())
         if system_inverter:
             requests.append(self.current_system_inverter_data())
+        if system_storage:
+            requests.append(self.current_system_storage_data())
         for i in device_meter:
             requests.append(self.current_meter_data(i))
         for i in device_storage:
@@ -220,7 +294,15 @@ class Fronius:
         for i in device_inverter:
             requests.append(self.current_inverter_data(i))
 
-        responses = await asyncio.gather(*requests, loop=loop)
+        res = await asyncio.gather(*requests, loop=loop, return_exceptions=True)
+        responses = []
+        for result in res:
+            if isinstance(result, FroniusError):
+                _LOGGER.warning(result)
+                if isinstance(result, BadStatusError):
+                    responses.append(result.response)
+                continue
+            responses.append(result)
         return responses
 
     @staticmethod
@@ -253,7 +335,7 @@ class Fronius:
         sensor = {}
         try:
             res = await self._fetch_solar_api(spec, spec_name, *spec_formattings)
-        except ValueError:
+        except InvalidAnswerError:
             # except if Host returns 404
             raise NotSupportedError(
                 "Device type {} not supported by the fronius device".format(spec_name)
@@ -262,10 +344,15 @@ class Fronius:
         try:
             sensor.update(Fronius._status_data(res))
         except (TypeError, KeyError):
-            # break if Data is empty
-            _LOGGER.info(
+            raise InvalidAnswerError(
                 "No header data returned from {} ({})".format(spec, spec_formattings)
             )
+        else:
+            if sensor["status"]["Code"] != 0:
+                endpoint = spec[self.api_version]
+                code = sensor["status"]["Code"]
+                reason = sensor["status"]["Reason"]
+                raise BadStatusError(endpoint, code, reason=reason, response=sensor)
         try:
             sensor.update(fun(res["Body"]["Data"]))
         except (TypeError, KeyError):
@@ -273,8 +360,7 @@ class Fronius:
             try:
                 sensor.update(fun(res["Body"]["LoggerInfo"]))
             except (TypeError, KeyError):
-                # break if Data is empty
-                _LOGGER.info(
+                raise InvalidAnswerError(
                     "No body data returned from {} ({})".format(spec, spec_formattings)
                 )
         return sensor
@@ -306,7 +392,7 @@ class Fronius:
             "current system inverter",
         )
 
-    async def current_meter_data(self, device=0):
+    async def current_meter_data(self, device: str = "0") -> Dict[str, Any]:
         """
         Get the current meter data for a device.
         """
@@ -314,7 +400,7 @@ class Fronius:
             Fronius._device_meter_data, URL_DEVICE_METER, "current meter", device
         )
 
-    async def current_storage_data(self, device=0):
+    async def current_storage_data(self, device: str = "0") -> Dict[str, Any]:
         """
         Get the current storage data for a device.
         Provides data about batteries.
@@ -329,10 +415,10 @@ class Fronius:
         Provides data about batteries.
         """
         return await self._current_data(
-            Fronius._system_storage_data, URL_DEVICE_STORAGE, "current system storage"
+            Fronius._system_storage_data, URL_SYSTEM_STORAGE, "current system storage"
         )
 
-    async def current_inverter_data(self, device=1):
+    async def current_inverter_data(self, device: str = "1") -> Dict[str, Any]:
         """
         Get the current inverter data of one device.
         """
@@ -369,6 +455,14 @@ class Fronius:
             Fronius._logger_info, URL_LOGGER_INFO, "current logger info"
         )
 
+    async def inverter_info(self):
+        """
+        Get the general infos of an inverter.
+        """
+        return await self._current_data(
+            Fronius._inverter_info, URL_INVERTER_INFO, "inverter info"
+        )
+
     @staticmethod
     def _system_led_data(data):
         _LOGGER.debug("Converting system led data: '{}'".format(data))
@@ -993,6 +1087,39 @@ class Fronius:
 
         return sensor
 
+    @staticmethod
+    def _inverter_info(data):
+        """Parse inverter info."""
+        _LOGGER.debug("Converting inverter info: '{}'".format(data))
+        inverters = []
+        for inverter_index, inverter_info in data.items():
+            inverter = {
+                "device_id": {"value": inverter_index},
+                "device_type": {"value": inverter_info["DT"]},
+                "pv_power": {"value": inverter_info["PVPower"], "unit": WATT},
+                "status_code": {"value": inverter_info["StatusCode"]},
+                "unique_id": {"value": inverter_info["UniqueID"]},
+            }
+            if inverter_info["DT"] in INVERTER_DEVICE_TYPE:
+                # add manufacturer and model if known
+                inverter["device_type"].update(
+                    INVERTER_DEVICE_TYPE[inverter_info["DT"]]
+                )
+            # "CustomName" not available on API V0 so default to ""
+            # html escaped by V1 Snap-In, UTF-8 by V1 Gen24
+            if "CustomName" in inverter_info:
+                inverter["custom_name"] = {
+                    "value": unescape(inverter_info["CustomName"])
+                }
+            # "ErrorCode" not in V1-Gen24
+            if "ErrorCode" in inverter_info:
+                inverter["error_code"] = {"value": inverter_info["ErrorCode"]}
+            # "Show" not in V0
+            if "Show" in inverter_info:
+                inverter["show"] = {"value": inverter_info["Show"]}
+            inverters.append(inverter)
+        return {"inverters": inverters}
+
     @staticmethod
     def _logger_info(data):
         _LOGGER.debug("Converting Logger info: '{}'".format(data))
